@@ -1,10 +1,12 @@
-# Version 1.2-1
+# Version 1.3-1
 
 import re
 import sys
 from hostlist import expand_hostlist, collect_hostlist
 import commands
 
+# Exceptions
+
 class NodesException(Exception):
     def __init__(self, msg):
         self.msg = msg
@@ -17,6 +19,13 @@ class UnknownLevel(NodesException): pass
 class AmbiguousLevel(NodesException): pass
 class FailedSLURM(NodesException): pass
 
+# Helper functions
+
+def collect_name_hostlist(node_iterable):
+    return collect_hostlist([node.name for node in node_iterable])
+
+# Classes
+
 class Entity:
     pass
 
@@ -28,7 +37,7 @@ class Node(Entity):
     def __repr__(self):
         return "<Node: %s>" % (self.name)
 
-    def expand_to_nodes(self):
+    def expand_to_node_set(self):
         return set([self])
 
     def add_to_group(self, group):
@@ -49,24 +58,24 @@ class Group(Entity):
         self.hierarchy = hierarchy
         self.level = level
         self.name = name
-        self.nodes = set() # Node set
+        self.node_set = set() # Node set
         self.all_nodes_group = all_nodes_group
 
     def __repr__(self):
         return "<Group %s: %s>" % (self.level, self.name)
 
-    def expand_to_nodes(self):
-        return self.nodes
+    def expand_to_node_set(self):
+        return self.node_set
 
-    def add_nodes(self, name_string):
-        #print "  NODES", self.name, "-->", name_string
+    def add_nodespec(self, nodespec):
+        #print "  NODES", self.name, "-->", nodespec
         node_set = set()
-        for name in expand_hostlist(name_string):
+        for name in expand_hostlist(nodespec):
             entity = self.hierarchy.get_entity(name, auto_add_node = self.all_nodes_group)
-            nodes = entity.expand_to_nodes()
+            nodes = entity.expand_to_node_set()
             node_set |= nodes
         #print "  NODE_SET", node_set
-        self.nodes |= node_set
+        self.node_set |= node_set
         for node in node_set:
             node.add_to_group(self)
 
@@ -98,8 +107,11 @@ class Hierarchy:
                 raise UnknownName("unknown name " + name)
         return self.names[name]
 
-    def parse_file(self, filename):
-        f = open(filename)
+    def parse_file(self, file_or_filename):
+        if type(file_or_filename) == type(""):
+            f = open(file_or_filename)
+        else:
+            f = file_or_filename
         current_group = None
         for line in f:
             line = re.sub(r' *#.*', '', line)
@@ -112,7 +124,7 @@ class Hierarchy:
                 rest = m.group(1)
                 current_group = self.create_group("all", "nodes", all_nodes_group = True)
                 if rest:
-                    current_group.add_nodes(rest)
+                    current_group.add_nodespec(rest)
                 continue
             
             # <level> <name>: <parts>
@@ -121,7 +133,7 @@ class Hierarchy:
                 (level, name, rest) = m.group(1,2,3)
                 current_group = self.create_group(level, name)
                 if rest:
-                    current_group.add_nodes(rest)
+                    current_group.add_nodespec(rest)
                 continue
 
             # <parts> (indented)
@@ -129,7 +141,7 @@ class Hierarchy:
                 m = re.match(r'^\s+(.*)', line)
                 if m: 
                     rest = m.group(1)
-                    current_group.add_nodes(rest)
+                    current_group.add_nodespec(rest)
                     continue
 
             # Fail
@@ -147,10 +159,10 @@ class Hierarchy:
             job, user, nodes = fields
 
             gj = self.create_group("job", "job" + job)
-            gj.add_nodes(nodes)
+            gj.add_nodespec(nodes)
 
             gu = self.create_or_get_group("user", user)
-            gu.add_nodes(nodes)
+            gu.add_nodespec(nodes)
 
     def expand_abbreviated_level(self, level):
         matching = [l for l in self.level_names if l.startswith(level)]
@@ -162,67 +174,82 @@ class Hierarchy:
             raise AmbiguousLevel("ambiguous level %s matching %s" % (level,
                                                                      ", ".join(matching)))
 
-    def expand_to_node_set(self, name_string):
+    def expand_to_node_set(self, nodespec):
         res = set()
-        for name in expand_hostlist(name_string):
+        for name in expand_hostlist(nodespec):
             entity = self.get_entity(name)
-            res |= entity.expand_to_nodes()
+            res |= entity.expand_to_node_set()
         return res
 
-    def expand_to_nodes(self, name_string):
-        return collect_hostlist([node.name for node in self.expand_to_node_set(name_string)])
+    # This is the method invoked by "nodes --to-nodes" (default) 
+    def expand(self, nodespec):
+        return collect_name_hostlist(self.expand_to_node_set(nodespec))
 
-    def up_set(self, level, name_string):
-        res = set()
-        missing = set()
-        for node in self.expand_to_node_set(name_string):
-            group = node.get_group(level)
-            if group is None:
-                missing.add(node)
-            else:
-                res.add(group)
-        return res, missing
+    # This is the central logic behind up, fill and gather
+    # Returns a tuple of sets (groups, leftovers, missing):
+    #   groups:    groups that nodes in nodespec belong to
+    #   leftovers: nodes that did not fill whole groups (or empty if fill is True)
+    #   missing:   nodes that did not belong to any group
 
-    def up(self, level, name_string):
-        level = self.expand_abbreviated_level(level)
-        groups, missing = self.up_set(level, name_string)
-        if missing:
-            raise MissingGroup("missing %s for %s" % \
-                                   (level,
-                                    collect_hostlist([node.name for node in missing])))
-
-        return collect_hostlist([group.name for group in groups])
-
-
-    def gather(self, level, name_string):
-        level = self.expand_abbreviated_level(level)
+    def up_set(self, level, nodespec, fill = True):
         groups = set()
+        leftovers = set()
         missing = set()
-        not_groupable = set()
-        node_set = self.expand_to_node_set(name_string)
+        node_set = self.expand_to_node_set(nodespec)
         for node in node_set:
             group = node.get_group(level)
             if group is None:
                 missing.add(node)
-            elif group.expand_to_nodes() <= node_set:
+            elif fill or group.expand_to_node_set() <= node_set:
                 groups.add(group)
             else:
-                not_groupable.add(node)
+                leftovers.add(node)
+
+        return groups, leftovers, missing
+
+    # This is the method invoked by "nodes --up"
+    def up(self, level, nodespec):
+        level = self.expand_abbreviated_level(level)
+        groups, leftovers, missing = self.up_set(level, nodespec)
+
+        assert len(leftovers) == 0
         if missing:
             raise MissingGroup("missing %s for %s" % \
                                    (level,
-                                    collect_hostlist([node.name for node in missing])))
+                                    collect_name_hostlist(missing)))
 
-        return collect_hostlist([group.name for group in groups|not_groupable])
+        return collect_name_hostlist(groups)
 
-    def fill(self, level, name_string):
+
+    # This is the method invoked by "nodes --fill"
+    def fill(self, level, nodespec):
         level = self.expand_abbreviated_level(level)
-        groups, res = self.up_set(level, name_string)
-        if res:
+        groups, leftovers, missing = self.up_set(level, nodespec)
+
+        assert len(leftovers) == 0
+        if missing:
             sys.stderr.write("Warning: missing %s for %s\n" % \
                                  (level,
-                                  collect_hostlist([node.name for node in res])))
+                                  collect_name_hostlist(missing)))
+
+        # FIXME: Check if it makes more sense to signal error above?
+        res = missing
+
         for group in groups:
-            res |= group.expand_to_nodes()
-        return collect_hostlist([node.name for node in res])
+            res |= group.expand_to_node_set()
+        return collect_name_hostlist(res)
+
+    # This is the method invoked by "nodes --gather"
+    def gather(self, level, nodespec):
+        level = self.expand_abbreviated_level(level)
+        groups, leftovers, missing = self.up_set(level, nodespec, fill=False)
+
+        if missing:
+            sys.stderr.write("Warning: missing %s for %s\n" % \
+                                 (level,
+                                  collect_name_hostlist(missing)))
+
+        return collect_name_hostlist(groups|leftovers|missing)
+
+
 
